import argparse
import datetime
import logging
import os

import h5py
from pynwb import NWBHDF5IO
from temporaldata import Data, IrregularTimeSeries, Interval

from brainsets.descriptions import (
    BrainsetDescription,
    SessionDescription,
    DeviceDescription,
)
from brainsets.utils.dandi_utils import (
    extract_spikes_from_nwbfile,
    extract_subject_from_nwb,
)
from brainsets.taxonomy import RecordingTech, Task
from brainsets import serialize_fn_map


logging.basicConfig(level=logging.INFO)


def extract_trials(nwbfile):
    r"""Extract trial information from the NWB file. Trials that are flagged as
    "to discard" or where the monkey failed are marked as invalid."""
    trial_table = nwbfile.trials.to_dataframe()

    # rename start and end time columns
    trial_table = trial_table.rename(
        columns={
            "start_time": "start",
            "stop_time": "end",
            "split": "split_indicator",
        }
    )
    trials = Interval.from_dataframe(trial_table)

    # the dataset has pre-defined train/valid splits, we will use the valid split
    # as our test
    train_mask_nwb = trial_table.split_indicator.to_numpy() == "train"
    test_mask_nwb = trial_table.split_indicator.to_numpy() == "val"

    trials.train_mask_nwb = (
        train_mask_nwb  # Naming with "_" since train_mask is reserved
    )
    trials.test_mask_nwb = test_mask_nwb  # Naming with "_" since test_mask is reserved

    return trials


def extract_behavior(nwbfile, trials):
    """Extract behavior from the NWB file.

    ..note::
        Cursor position and target position are in the same frame of reference.
        They are both of size (sequence_len, 2). Finger position can be either 3d or 6d,
        depending on the sequence. # todo investigate more
    """
    # cursor, hand and eye share the same timestamps (verified)
    timestamps = nwbfile.processing["behavior"]["hand_vel"].timestamps[:]
    hand_pos = nwbfile.processing["behavior"]["hand_pos"].data[:]
    hand_vel = nwbfile.processing["behavior"]["hand_vel"].data[:]
    eye_pos = nwbfile.processing["behavior"]["eye_pos"].data[:]

    hand = IrregularTimeSeries(
        timestamps=timestamps,
        pos=hand_pos,
        vel=hand_vel,
        domain="auto",
    )

    eye = IrregularTimeSeries(
        timestamps=timestamps,
        pos=eye_pos,
        domain="auto",
    )

    return hand, eye


def main():
    # use argparse to get arguments from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_file", type=str)
    parser.add_argument("--output_dir", type=str, default="./processed")

    args = parser.parse_args()

    # intiantiate a DatasetBuilder which provides utilities for processing data
    brainset_description = BrainsetDescription(
        id="pei_pandarinath_nlb_2021",
        origin_version="dandi/000140/0.220113.0408",
        derived_version="1.0.0",
        source="https://dandiarchive.org/dandiset/000140",
        description="This dataset contains sorted unit spiking times and behavioral"
        " data from a macaque performing a delayed reaching task. The experimental task"
        " was a center-out reaching task with obstructing barriers forming a maze,"
        " resulting in a variety of straight and curved reaches.",
    )

    logging.info(f"Processing file: {args.input_file}")

    # open file
    io = NWBHDF5IO(args.input_file, "r")
    nwbfile = io.read()

    # extract subject metadata
    # this dataset is from dandi, which has structured subject metadata, so we
    # can use the helper function extract_subject_from_nwb
    subject = extract_subject_from_nwb(nwbfile)

    # extract experiment metadata
    recording_date = nwbfile.session_start_time.strftime("%Y%m%d")
    device_id = f"{subject.id}_{recording_date}"
    session_id = f"{subject.id}_maze"

    # register session
    session_description = SessionDescription(
        id=session_id,
        recording_date=datetime.datetime.strptime(recording_date, "%Y%m%d"),
        task=Task.REACHING,
    )

    # register device
    device_description = DeviceDescription(
        id=device_id,
        recording_tech=RecordingTech.UTAH_ARRAY_SPIKES,
    )

    # extract spiking activity
    # this data is from dandi, we can use our helper function
    spikes, units = extract_spikes_from_nwbfile(
        nwbfile,
        recording_tech=RecordingTech.UTAH_ARRAY_SPIKES,
    )

    # extract data about trial structure
    trials = extract_trials(nwbfile)

    data = Data(
        brainset=brainset_description,
        session=session_description,
        device=device_description,
        # neural activity
        spikes=spikes,
        units=units,
        # stimuli and behavior
        trials=trials,
        # domain
        domain="auto",
    )

    if not "test" in args.input_file:
        # extract behavior
        data.hand, data.eye = extract_behavior(nwbfile, trials)

        # report accuracy only on the evaluation intervals
        data.nlb_eval_intervals = Interval(
            start=trials.move_onset_time - 0.05,
            end=trials.move_onset_time + 0.65,
        )

        # split and register trials into train, validation and test
        train_trials, valid_trials = trials.select_by_mask(trials.train_mask_nwb).split(
            [0.8, 0.2], shuffle=True, random_seed=42
        )
        test_trials = trials.select_by_mask(trials.test_mask_nwb)

        data.set_train_domain(train_trials)
        data.set_valid_domain(valid_trials)
        data.set_test_domain(test_trials)

    # close file
    io.close()

    # save data to disk
    path = os.path.join(args.output_dir, f"{session_id}.h5")
    with h5py.File(path, "w") as file:
        data.to_hdf5(file, serialize_fn_map=serialize_fn_map)


if __name__ == "__main__":
    main()



from typing import Optional
import click
from pathlib import Path
from prompt_toolkit import prompt
from prompt_toolkit.shortcuts import CompleteStyle
from prompt_toolkit.completion import PathCompleter

from .utils import load_config, save_config, expand_path


@click.command()
@click.option(
    "--raw-dir",
    help="Path for storing raw data.",
    type=click.Path(file_okay=False, dir_okay=True),
    required=False,
)
@click.option(
    "--processed-dir",
    help="Path for storing processed brainsets.",
    type=click.Path(file_okay=False, dir_okay=True),
    required=False,
)
def config(raw_dir: Optional[Path], processed_dir: Optional[Path]):
    """Set raw and processed data directories."""

    # Get missing args from user prompts
    if raw_dir is None or processed_dir is None:
        raw_dir = prompt(
            "Enter raw data directory: ",
            completer=PathCompleter(only_directories=True),
            complete_style=CompleteStyle.READLINE_LIKE,
        )
        processed_dir = prompt(
            "Enter processed data directory: ",
            completer=PathCompleter(only_directories=True),
            complete_style=CompleteStyle.READLINE_LIKE,
        )

    raw_dir = expand_path(raw_dir)
    processed_dir = expand_path(processed_dir)

    # Create directories
    raw_dir.mkdir(parents=True, exist_ok=True)
    processed_dir.mkdir(parents=True, exist_ok=True)

    # Save config
    config = load_config(raise_cli_error=False)
    config_exists = config is not None
    if not config_exists:
        config = {}
    config["raw_dir"] = str(raw_dir)
    config["processed_dir"] = str(processed_dir)
    config_filepath = save_config(config)

    if not config_exists:
        click.echo(f"Created config file at {config_filepath}")
    else:
        click.echo(f"Updated config file at {config_filepath}")

    click.echo(f"Raw data dir: {config['raw_dir']}")
    click.echo(f"Processed data dir: {config['processed_dir']}")



from brainsets.core import StringIntEnum


# Nomenclature from CHARM: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2944191/
# These areas are defined at different levels of precision - have a look at the CHARM_Key_table
# to find how these
class Macaque(StringIntEnum):
    Frontal_Lobe = 1
    anterior_cingulate_gyrus = 2
    anterior_cingulate_cortex = 3
    area_32 = 4
    subgenual_cortex = 5
    area_24 = 6
    areas_24a_and_24b = 7
    area_24a = 8
    area_24b = 9
    area_24c = 10
    midcingulate_cortex = 11
    areas_24a_and_24b_prime = 12
    area_24a_prime = 13
    area_24b_prime = 14
    area_24c_prime = 15
    orbital_frontal_cortex = 16
    medial_orbital_frontal_cortex = 17
    frontal_pole = 18
    rostral_medial_frontal_pole = 19
    caudal_medial_frontal_pole = 20
    orbital_frontal_pole = 21
    gyrus_rectus = 22
    rostral_gyrus_rectus = 23
    caudal_gyrus_rectus = 24
    lateral_orbital_frontal_cortex = 25
    areas_11_and_13 = 26
    area_11 = 27
    medial_area_11 = 28
    lateral_area_11 = 29
    area_13 = 30
    areas_13a_and_13b = 31
    medial_area_13 = 32
    lateral_area_13 = 33
    medial_and_orbital_area_12 = 34
    medial_area_12 = 35
    orbital_area_12 = 36
    caudal_orbital_frontal_cortex = 37
    agranular_orbitofrontal_cortex = 38
    medial_agranular_insular_region = 39
    lateral_agranular_insular_region = 40
    intermediate_agranular_insula_area = 41
    lateral_agranular_insula_area = 42
    posterior_lateral_agranular_insula_area = 43
    primary_olfactory_cortex = 44
    anterior_olfactory_cortex = 45
    posterior_olfactory_cortex = 46
    caudolateral_orbitofrontal_cortex = 47
    gustatory_cortex = 48
    precentral_operular_area = 49
    lateral_prefrontal_cortex = 50
    periarcuate_area_8A_Frontal_Eye_Fields = 51
    dorsal_periarcuate_area_8A = 52
    ventral_periarcuate_area_8A = 53
    dorsolateral_prefrontal_cortex = 54
    area_8B = 55
    dorsal_area_8B = 56
    medial_area_8B = 57
    arcuate_sulcus_area_8B = 58
    rostral_dorsolateral_prefrontal_cortex = 59
    area_9 = 60
    dorsal_area_9 = 61
    medial_area_9 = 62
    dorsal_area_46 = 63
    ventrolateral_prefrontal_cortex = 64
    rostral_ventrolateral_prefrontal_cortex = 65
    ventral_and_fundus_portion_of_area_46 = 66
    area_46_in_fundus_of_the_principal_sulcus = 67
    ventral_area_46 = 68
    rostral_and_lateral_area_12 = 69
    lateral_area_12 = 70
    rostral_area_12 = 71
    caudal_ventrolateral_prefrontal_cortex = 72
    area_45 = 73
    area_45a = 74
    area_45b = 75
    area_44 = 76
    motor_cortex = 77
    lateral_motor_cortex = 78
    primary_motor_cortex = 79
    premotor_cortex = 80
    dorsal_premotor_cortex = 81
    caudal_dorsal_premotor_cortex = 82
    rostral_dorsal_premotor_cortex = 83
    ventral_premotor_cortex = 84
    area_F4_of_ventral_premotor_cortex = 85
    area_F5_of_ventral_premotor_cortex = 86
    medial_supplementary_motor_areas = 87
    presupplementary_motor_area = 88
    supplementary_motor_area = 89
    Parietal_Lobe = 90
    somatosensory_cortex = 91
    primary_somatosensory_cortex = 92
    areas_3a_and_3b = 93
    areas_1_2 = 94
    secondary_somatosensory_cortex = 95
    superior_parietal_lobule = 96
    visual_areas_V6_and_V6A = 97
    visual_area_V6 = 98
    visual_area_V6A = 99
    ventral_visual_area_6A = 100
    dorsal_visual_area_6A = 101
    area_5 = 102
    dorsal_area_5 = 103
    areas_PEc_and_PEci = 104
    area_PE = 105
    area_PEa = 106
    ventromedial_intraparietal_sulcus = 107
    medial_intraparietal_area = 108
    fundus_of_the_intraparietal_sulcus = 109
    ventral_intraparietal_area = 110
    posterior_intraparietal_area = 111
    inferior_parietal_lobule = 112
    lateral_intraparietal_sulcus = 113
    anterior_intraparietal_area = 114
    lateral_intraparietal_area = 115
    dorsal_lateral_intraparietal_area = 116
    ventral_lateral_intraparietal_area = 117
    lateral_occipital_parietal_area = 118
    medial_superior_temporal_area = 119
    area_7_in_the_inferior_parietal_lobule = 120
    inferior_parietal_lobule_areas_7a_and_7b = 121
    caudal_inferior_parietal_lobule_area_7a__Opt_PG = 122
    rostral_inferior_parietal_lobule_area_7b_PFG_PF = 123
    parietal_operculum = 124
    posterior_medial_cortex = 125
    area_7_PGm_on_the_medial_wall = 126
    posterior_cingulate_gyrus = 127
    posterior_cingulate_cortex = 128
    area_31 = 129
    area_23 = 130
    area_23a = 131
    area_23b = 132
    area_23c = 133
    area_v23 = 134
    retrosplenial_cortex = 135
    area_29 = 136
    area_30 = 137
    Temporal_Lobe = 138
    medial_temporal_lobe = 139
    hippocampus = 140
    hippocampal_formation = 141
    hippocampus_proper = 142
    hippocampal_subfields_CA1_CA3 = 143
    hippocampal_subfield_CA4 = 144
    dentate_gyrus = 145
    subiculum_and_prosubiculum = 146
    presubiculum_and_parasubiculum = 147
    parahippocampal_cortex = 148
    areas_TF_and_TFO = 149
    area_TFO = 150
    area_TF = 151
    area_TH = 152
    rhinal_cortex = 153
    entorhinal_cortex = 154
    caudal_entorhinal_cortex = 155
    caudal_division_of_entorhinal_cortex = 156
    caudal_limiting_division_of_entorhinal_cortex = 157
    middle_entorhinal_cortex = 158
    rostral_entorhinal_cortex = 159
    perirhinal_cortex = 160
    area_35 = 161
    area_36 = 162
    caudal_area_36 = 163
    rostral_and_polar_area_36 = 164
    temporal_pole = 165
    agranular_temporal_pole = 166
    granular_and_dysgranular_temporal_pole = 167
    dysgranular_temporal_pole = 168
    dorsal_dysgranular_temporal_pole = 169
    ventral_dysgranular_temporal_pole = 170
    granular_temporal_pole = 171
    dorsal_granular_temporal_pole = 172
    ventral_granular_temporal_pole = 173
    sts_part_of_temporal_pole = 174
    inferior_temporal_cortex = 175
    area_TEO = 176
    area_TE = 177
    gyral_aspect_of_area_TE = 178
    posterior_area_TE = 179
    posterior_ventral_area_TE = 180
    posterior_dorsal_area_TE = 181
    anterior_area_TE = 182
    anterior_ventral_area_TE = 183
    anterior_dorsal_area_TE = 184
    area_TE_in_the_ventral_STS = 185
    area_TEm = 186
    area_TEa = 187
    fundus_of_the_superior_temporal_sulcus = 188
    anterior_fundus_of_the_superior_temporal_sulcus = 189
    area_IPa = 190
    area_PGa = 191
    floor_of_the_superior_temporal_area = 192
    superior_temporal_region = 193
    rostral_superior_temporal_region = 194
    dorsal_bank_of_the_superior_temporal_sulcus = 195
    temporal_parietooccipital_associated_area = 196
    area_TAa = 197
    rostral_superior_temporal_gyrus = 198
    caudal_superior_temporal_gyrus = 199
    temporo_parietal_area = 200
    parabelt_areas_of_auditory_cortex = 201
    caudal_parabelt_region = 202
    rostral_parabelt_region = 203
    core_and_belt_areas_of_auditory_cortex = 204
    belt_areas_of_auditory_cortex = 205
    lateral_belt_areas = 206
    caudal_areas_of_the_lateral_belt = 207
    caudal_lateral_belt_region = 208
    middle_lateral_belt_region = 209
    rostral_areas_of_the_lateral_belt = 210
    anterior_lateral_belt_region = 211
    lateral_rostrotemporal_belt_region = 212
    medial_belt_areas = 213
    caudomedial_belt_region = 214
    rostral_areas_of_the_medial_belt = 215
    rostromedial_belt_region = 216
    medial_rostrotemporal_belt_region = 217
    polar_rostrotemporal_cortex = 218
    core_areas_of_auditory_cortex = 219
    rostral_areas_of_the_core = 220
    rostral_core_region = 221
    rostrotemporal_core_region = 222
    primary_auditory_cortex = 223
    floor_of_the_lateral_sulcus = 224
    insula_and_parainsula = 225
    parainsula = 226
    insula = 227
    agranular_and_dysgranular_insula = 228
    granular_insula = 229
    retroinsula = 230
    Occipital_Lobe = 231
    middle_temporal_area = 232
    extrastriate_visual_areas_2_4 = 233
    visual_area_4 = 234
    dorsal_visual_area_4 = 235
    ventral_visual_area_4 = 236
    preoccipital_visual_areas_2_3 = 237
    visual_area_3 = 238
    visual_areas_V3d_and_V3A = 239
    visual_area_V3A = 240
    dorsal_visual_area_3 = 241
    ventral_visual_area_3 = 242
    visual_area_2 = 243
    possibly_part_of_visual_area_2 = 244
    clearly_part_of_visual_area_2 = 245
    primary_visual_cortex = 246



import pytest
import inspect
from brainsets.core import StringIntEnum
import brainsets.taxonomy


def find_duplicates(enum):
    # Collect all enum names and their values
    enum_items = [(name, member.value) for name, member in enum.__members__.items()]

    # Find duplicates: create a dictionary where each value points to a list of names that share it
    value_to_names = {}
    for name, value in enum_items:
        if value in value_to_names:
            value_to_names[value].append(name)
        else:
            value_to_names[value] = [name]

    # Filter out entries with only one name, leaving only duplicates
    duplicates = {}
    for value, names in value_to_names.items():
        if len(names) > 1:
            # Check if names are allowed aliases (e.g., A = SOME_ALIAS = 0)
            if len(set(getattr(enum, name) for name in names)) > 1:
                duplicates[value] = names

    return duplicates


def get_all_stringintenum_subclasses():
    classes = []
    for module_name, module in inspect.getmembers(brainsets.taxonomy, inspect.ismodule):
        for name, obj in inspect.getmembers(module):
            if (
                inspect.isclass(obj)
                and issubclass(obj, StringIntEnum)
                and obj != StringIntEnum
            ):
                classes.append((name, obj))
    return classes


def test_all_stringintenum_classes_have_no_duplicate_ids():
    for name, cls in get_all_stringintenum_subclasses():
        duplicates = find_duplicates(cls)
        assert len(duplicates) == 0, f"Duplicate IDs found in {name} enum: {duplicates}"



import logging
import numpy as np
from temporaldata import Interval


def split_one_epoch(epoch, grid, split_ratios=[0.6, 0.1, 0.3]):
    assert len(epoch) == 1
    epoch_start = epoch.start[0]
    epoch_end = epoch.end[0]

    train_val_split_time = epoch_start + split_ratios[0] * (epoch_end - epoch_start)
    val_test_split_time = train_val_split_time + split_ratios[1] * (
        epoch_end - epoch_start
    )

    grid_match = grid.slice(
        train_val_split_time, train_val_split_time, reset_origin=False
    )
    if len(grid_match) > 0:
        if (
            train_val_split_time - grid_match.start[0]
            > grid_match.end[0] - train_val_split_time
        ):
            train_val_split_time = grid_match.end[0]
        else:
            train_val_split_time = grid_match.start[0]

    grid_match = grid.slice(
        val_test_split_time, val_test_split_time, reset_origin=False
    )
    if len(grid_match) > 0:
        if (
            val_test_split_time - grid_match.start[0]
            > grid_match.end[0] - val_test_split_time
        ):
            val_test_split_time = grid_match.end[0]
        else:
            val_test_split_time = grid_match.start[0]

    train_interval = Interval(start=epoch_start, end=train_val_split_time)
    val_interval = Interval(start=train_interval.end[0], end=val_test_split_time)
    test_interval = Interval(start=val_interval.end[0], end=epoch_end)

    return train_interval, val_interval, test_interval


def split_two_epochs(epoch, grid):
    assert len(epoch) == 2
    first_epoch_start = epoch.start[0]
    first_epoch_end = epoch.end[0]

    split_time = first_epoch_start + 0.5 * (first_epoch_end - first_epoch_start)
    grid_match = grid.slice(split_time, split_time, reset_origin=False)
    if len(grid_match) > 0:
        if split_time - grid_match.start[0] > grid_match.end[0] - split_time:
            split_time = grid_match.end[0]
        else:
            split_time = grid_match.start[0]

    train_interval = Interval(
        start=first_epoch_start,
        end=split_time,
    )
    val_interval = Interval(start=train_interval.end[0], end=first_epoch_end)
    test_interval = epoch.select_by_mask(np.array([False, True]))

    return train_interval, val_interval, test_interval


def split_three_epochs(epoch, grid):
    assert len(epoch) == 3

    test_interval = epoch.select_by_mask(np.array([False, False, True]))
    train_interval = epoch.select_by_mask(np.array([True, True, False]))

    split_time = train_interval.end[1] - 0.3 * (
        train_interval.end[1] - train_interval.start[1]
    )
    grid_match = grid.slice(split_time, split_time, reset_origin=False)
    if len(grid_match) > 0:
        if split_time - grid_match.start[0] > grid_match.end[0] - split_time:
            split_time = grid_match.end[0]
        else:
            split_time = grid_match.start[0]

    train_interval.end[1] = split_time
    val_interval = Interval(start=train_interval.end[1], end=epoch.end[1])

    return train_interval, val_interval, test_interval


def split_four_epochs(epoch, grid):
    assert len(epoch) == 4

    test_interval = epoch.select_by_mask(np.array([False, False, False, True]))
    train_interval = epoch.select_by_mask(np.array([True, True, True, False]))
    split_time = train_interval.end[2] - 0.5 * (
        train_interval.end[2] - train_interval.start[2]
    )
    grid_match = grid.slice(split_time, split_time, reset_origin=False)
    if len(grid_match) > 0:
        if split_time - grid_match.start[0] > grid_match.end[0] - split_time:
            split_time = grid_match.end[0]
        else:
            split_time = grid_match.start[0]

    train_interval.end[2] = split_time
    val_interval = Interval(start=train_interval.end[2], end=epoch.end[2])

    return train_interval, val_interval, test_interval


def split_five_epochs(epoch, grid):
    assert len(epoch) == 5

    train_interval = epoch.select_by_mask(np.array([True, True, True, False, False]))
    test_interval = epoch.select_by_mask(np.array([False, False, False, True, True]))

    split_time = train_interval.end[2] - 0.5 * (
        train_interval.end[2] - train_interval.start[2]
    )
    grid_match = grid.slice(split_time, split_time, reset_origin=False)
    if len(grid_match) > 0:
        if split_time - grid_match.start[0] > grid_match.end[0] - split_time:
            split_time = grid_match.end[0]
        else:
            split_time = grid_match.start[0]

    train_interval.end[2] = split_time
    val_interval = Interval(start=train_interval.end[2], end=epoch.end[2])

    return train_interval, val_interval, test_interval


def split_more_than_five_epochs(epoch):
    assert len(epoch) > 5

    train_interval, val_interval, test_interval = epoch.split(
        [0.6, 0.1, 0.3], shuffle=False
    )
    return train_interval, val_interval, test_interval


def generate_train_valid_test_splits(epoch_dict, grid):
    train_intervals = Interval(np.array([]), np.array([]))
    valid_intervals = Interval(np.array([]), np.array([]))
    test_intervals = Interval(np.array([]), np.array([]))

    for name, epoch in epoch_dict.items():
        if name == "invalid_presentation_epochs":
            logging.warn(f"Found invalid presentation epochs, which will be excluded.")
            continue
        if len(epoch) == 1:
            train, valid, test = split_one_epoch(epoch, grid)
        elif len(epoch) == 2:
            train, valid, test = split_two_epochs(epoch, grid)
        elif len(epoch) == 3:
            train, valid, test = split_three_epochs(epoch, grid)
        elif len(epoch) == 4:
            train, valid, test = split_four_epochs(epoch, grid)
        elif len(epoch) == 5:
            train, valid, test = split_five_epochs(epoch, grid)
        else:
            train, valid, test = split_more_than_five_epochs(epoch)

        train_intervals = train_intervals | train
        valid_intervals = valid_intervals | valid
        test_intervals = test_intervals | test

    return train_intervals, valid_intervals, test_intervals



from brainsets.core import StringIntEnum


class Species(StringIntEnum):
    UNKNOWN = 0
    MACACA_MULATTA = 1
    HOMO_SAPIENS = HUMAN = 2
    MUS_MUSCULUS = 3
    MACACA_FASCICULARIS = NCBITaxon_9541 = 4
    CAENORHABDITIS_ELEGANS = 5


class Sex(StringIntEnum):
    r"""Follows the DANDI definition of sex.
    [Link](https://docs.dandiarchive.org/135_validation/#missing-dandi-metadata)
    """

    UNKNOWN = U = 0
    MALE = M = 1
    FEMALE = F = 2
    OTHER = O = 3



import datetime
from typing import Dict, List, Tuple, Optional, Union

from pydantic.dataclasses import dataclass
import temporaldata

import brainsets
from brainsets.taxonomy import *
from brainsets.taxonomy.mice import *


@dataclass
class BrainsetDescription(temporaldata.Data):
    r"""A class for describing a brainset.

    Parameters
    ----------
    id : str
        Unique identifier for the brainset
    origin_version : str
        Version identifier for the original data source
    derived_version : str
        Version identifier for the derived/processed data
    source : str
        Original data source (usually a URL, or a short description otherwise)
    description : str
        Text description of the brainset
    brainsets_version : str, optional
        Version of brainsets package used, defaults to current version
    temporaldata_version : str, optional
        Version of temporaldata package used, defaults to current version
    """

    id: str
    origin_version: str
    derived_version: str
    source: str
    description: str
    brainsets_version: str = brainsets.__version__
    temporaldata_version: str = temporaldata.__version__


@dataclass
class SubjectDescription(temporaldata.Data):
    r"""A class for describing a subject.

    Parameters
    ----------
    id : str
        Unique identifier for the subject
    species : Species
        Species of the subject
    age : float, optional
        Age of the subject in days, defaults to 0.0
    sex : Sex, optional
        Sex of the subject, defaults to UNKNOWN
    genotype : str, optional
        Genotype of the subject, defaults to "unknown"
    cre_line : Cre_line, optional
        Cre line of the subject, defaults to None
    """

    id: str
    species: Species
    age: float = 0.0  # in days
    sex: Sex = Sex.UNKNOWN
    genotype: str = "unknown"  # no idea how many there will be for now.
    cre_line: Optional[Cre_line] = None


@dataclass
class SessionDescription(temporaldata.Data):
    r"""A class for describing an experimental session.

    Parameters
    ----------
    id : str
        Unique identifier for the session
    recording_date : datetime.datetime
        Date and time when the recording was made
    task : Task
        Task performed during the session
    """

    id: str
    recording_date: datetime.datetime
    task: Optional[Task] = None


@dataclass
class DeviceDescription(temporaldata.Data):
    r"""A class for describing a recording device.

    Parameters
    ----------
    id : str
        Unique identifier for the device
    recording_tech : RecordingTech or List[RecordingTech], optional
        Recording technology used, defaults to None
    processing : str, optional
        Processing applied to the recording, defaults to None
    chronic : bool, optional
        Whether the device was chronically implanted, defaults to False
    start_date : datetime.datetime, optional
        Date when device was implanted/first used, defaults to None
    end_date : datetime.datetime, optional
        Date when device was removed/last used, defaults to None
    imaging_depth : float, optional
        Depth of imaging in micrometers, defaults to None
    target_area : BrainRegion, optional
        Target brain region for recording, defaults to None
    """

    id: str
    # units: List[str]
    # areas: Union[List[StringIntEnum], List[Macaque]]
    recording_tech: Union[RecordingTech, List[RecordingTech]] = None
    processing: Optional[str] = None
    chronic: bool = False
    start_date: Optional[datetime.datetime] = None
    end_date: Optional[datetime.datetime] = None
    # Ophys
    imaging_depth: Optional[float] = None  # in um
    target_area: Optional[BrainRegion] = None



import os
import pathlib
import shutil
from collections import defaultdict
import logging

from rich import print
from rich.filesize import decimal
from rich.markup import escape
from rich.prompt import Confirm, Prompt
from rich.text import Text
from rich.tree import Tree


def find_files_by_extension(folder_path, extension):
    for root, dirs, files in os.walk(folder_path):
        for file in files:
            if file.endswith(extension):
                yield os.path.join(root, file)


def make_directory(path, prompt_if_exists=False):
    if not os.path.exists(path):
        os.makedirs(path)
        logging.info("New directory [u]{}[/u] created.".format(path))
    else:
        logging.info("Directory [u]{}[/u] already exists.".format(path))
        if not prompt_if_exists:
            return

        rm_dir = Confirm.ask(
            "Do you want to delete the folder? "
            "[bold red blink]This will permanently remove its contents.[/bold red blink]",
            default=False,
        )

        if rm_dir:
            shutil.rmtree(path)
            os.makedirs(path)
            logging.info("New directory [u]{}[/u] created.".format(path))
        else:
            # kill process
            logging.info("Killing process.")
            exit()


def get_dir_tree(path):
    tree = Tree(
        f":open_file_folder: [link file://{path}]{path}",
        guide_style="bold bright_blue",
    )
    walk_directory(pathlib.Path(path), tree)
    return tree


def walk_directory(directory: pathlib.Path, tree: Tree) -> None:
    paths = sorted(
        pathlib.Path(directory).iterdir(),
        key=lambda path: (path.is_file(), path.name.lower()),
    )

    for path in paths:
        # Remove hidden files
        if path.name.startswith("."):
            continue

        if path.is_dir():
            style = "dim" if path.name.startswith("__") else ""
            branch = tree.add(
                f"[bold magenta]:open_file_folder: [link file://{path}]{escape(path.name)}",
                style=style,
                guide_style=style,
            )
            walk_directory(path, branch)
        else:
            text_filename = Text(path.name, "green")
            text_filename.highlight_regex(r"\..*$", "bold red")
            text_filename.stylize(f"link file://{path}")
            file_size = path.stat().st_size
            text_filename.append(f" ({decimal(file_size)})", "blue")
            try:
                create_time = path.stat().st_birthtime
                text_filename.append(f" ({create_time})", "blue")
            except:
                pass
            icon = defaultdict(
                lambda: "📄 ",
                py="🐍 ",
                cfg="🛠 ",
                pt="🔥 ",
                nwb="🧠 ",
                npy="🐍 ",
                mat="📊 ",
            )[path.suffix[1:]]
            tree.add(Text(icon) + text_filename)



import numpy as np
import pytest
from brainsets.processing.signal import (
    extract_bands,
    cube_to_long,
)


@pytest.mark.skipif(pytest.importorskip("mne") is None, reason="mne not installed")
def test_extract_bands():
    bands, ts, band_names = extract_bands(
        np.random.randn(1000, 10), np.arange(1000), Fs=1000, notch=60
    )
    assert bands.ndim == 3
    assert ts.size == bands.shape[0]
    assert bands.shape[2] == len(band_names)


@pytest.mark.skipif(pytest.importorskip("mne") is None, reason="mne not installed")
def test_cube_to_long():
    ts = np.arange(10).astype(float)
    cube = np.zeros((1, 10, 2), dtype=int)
    cube[0, 6, 0] = 2
    cube[0, 5, 1] = 3

    trials, units = cube_to_long(ts, cube)

    assert len(units.id) == 2
    assert len(trials) == 1
    assert len(trials[0].timestamps) == 5
    assert np.allclose(np.array([1, 1, 1, 0, 0]), trials[0].unit_index)
    assert np.allclose(np.array([5, 5, 5, 6, 6]), trials[0].timestamps)



import argparse
import datetime
import logging
import h5py
import os

import numpy as np
from pynwb import NWBHDF5IO
from scipy.ndimage import binary_dilation
import pandas as pd
from temporaldata import (
    Data,
    IrregularTimeSeries,
    Interval,
    ArrayDict,
)
from brainsets.descriptions import (
    BrainsetDescription,
    SessionDescription,
    DeviceDescription,
)
from brainsets.taxonomy import RecordingTech, Task
from brainsets.utils.dandi_utils import extract_subject_from_nwb
from brainsets import serialize_fn_map

logging.basicConfig(level=logging.INFO)


def extract_trials(nwbfile, session_id):
    r"""Extract trial information from the NWB file. Trials that are flagged as
    "to discard" or where the monkey failed are marked as invalid."""
    trial_table = nwbfile.trials.to_dataframe()
    # rename a few columns
    trial_table = trial_table.rename(
        columns={
            "start_time": "start",
            "stop_time": "end",
            "target_presentation_time": "target_on_time",
        }
    )
    timekeys = [
        "start",
        "end",
        "target_on_time",
        "go_cue_time",
        "move_begins_time",
        "move_ends_time",
    ]

    # some sessions' raw data is messed up, we need to fix it
    artifact_index = np.where(
        trial_table.start.to_numpy()[1:] < trial_table.end.to_numpy()[:-1]
    )[0]
    artifact_dict = {}
    if len(artifact_index) > 0:
        logging.info(f"Found {len(artifact_index)} artifacts in the trial table.")
        if session_id in [
            "nitschke_20090910_center_out_reaching",
            "nitschke_20090920_center_out_reaching",
        ]:
            # the first artifact corresponds to two overlapping trials
            # we will skip those two trials
            skip_trials = [artifact_index[0], artifact_index[0] + 1]
            logging.info(f"Dropping trials {skip_trials} due to artifacts.")
            artifact_dict["skip_start"] = trial_table.start.to_numpy()[skip_trials][0]
            artifact_dict["skip_end"] = trial_table.end.to_numpy()[skip_trials][1]
            trial_table = trial_table.drop(index=skip_trials).reset_index(drop=True)
            artifact_index = artifact_index[1:]
            artifact_index = artifact_index - 2

        epoch_start_indices = np.concatenate([np.array([0]), artifact_index + 1])
        epoch_end_indices = np.concatenate(
            [artifact_index + 1, np.array([len(trial_table)])]
        )

        # we split the trial table into epochs
        trial_table_list = [
            trial_table.iloc[start_idx:end_idx].copy()
            for start_idx, end_idx in zip(epoch_start_indices, epoch_end_indices)
        ]

        # we estimate the start and end of each epoch
        epoch_start_times = [
            trial_table_list[i].start.to_numpy()[0]
            for i in range(len(trial_table_list))
        ]
        epoch_end_times = [
            trial_table_list[i].end.to_numpy()[-1] for i in range(len(trial_table_list))
        ]

        artifact_dict["original_start_times"] = epoch_start_times
        artifact_dict["original_end_times"] = epoch_end_times
        artifact_dict["fixed_start_times"] = [epoch_start_times[0]]
        artifact_dict["fixed_end_times"] = [epoch_end_times[0]]
        gap = 60  # seconds
        for i, trial_block in enumerate(trial_table_list):
            if i == 0:
                # we keep the first epoch as is
                continue
            # the next epochs will be reset to start "gap" seconds after the previous epoch
            for key in timekeys:
                trial_block[key] = (
                    trial_block[key]
                    - artifact_dict["original_start_times"][i]
                    + artifact_dict["fixed_end_times"][i - 1]
                    + gap
                )
            artifact_dict["fixed_start_times"].append(
                epoch_start_times[i]
                - artifact_dict["original_start_times"][i]
                + artifact_dict["fixed_end_times"][i - 1]
                + gap
            )
            artifact_dict["fixed_end_times"].append(
                epoch_end_times[i]
                - artifact_dict["original_start_times"][i]
                + artifact_dict["fixed_end_times"][i - 1]
                + gap
            )

        # then we recombine everything!
        trial_table = pd.concat(trial_table_list)

    trials = Interval.from_dataframe(trial_table, timekeys=timekeys)

    assert trials.is_disjoint()
    assert trials.is_sorted()

    trials.is_valid = np.logical_and(
        trials.discard_trial == 0.0, trials.task_success == 1.0
    )
    valid_trials = trials.select_by_mask(trials.is_valid)

    movement_phases = Data(
        hold_period=Interval(
            start=valid_trials.target_on_time, end=valid_trials.go_cue_time
        ),
        reach_period=Interval(
            start=valid_trials.move_begins_time, end=valid_trials.move_ends_time
        ),
        return_period=Interval(start=valid_trials.move_ends_time, end=valid_trials.end),
        domain="auto",
    )

    return trials, movement_phases, artifact_dict


def extract_behavior(nwbfile, artifact_dict):
    """Extract behavior from the NWB file.

    ..note::
        Cursor position and target position are in the same frame of reference.
        They are both of size (sequence_len, 2). Finger position can be either 3d or 6d,
        depending on the sequence. # todo investigate more
    """
    # cursor, hand and eye share the same timestamps (verified)
    timestamps = nwbfile.processing["behavior"]["Position"]["Cursor"].timestamps[:]
    cursor_pos = nwbfile.processing["behavior"]["Position"]["Cursor"].data[:]  # 2d
    hand_pos = nwbfile.processing["behavior"]["Position"]["Hand"].data[:]
    eye_pos = nwbfile.processing["behavior"]["Position"]["Eye"].data[:]  # 2d

    # some sessions' raw data is messed up, we need to fix it
    artifact_index = np.where(timestamps[1:] < timestamps[:-1])[0]
    assert (len(artifact_index) > 0) == (len(artifact_dict) > 0)
    if len(artifact_dict) > 0:
        if "skip_start" in artifact_dict:
            mask = np.logical_and(
                timestamps >= artifact_dict["skip_start"],
                timestamps <= artifact_dict["skip_end"],
            )
            mask[artifact_index[0] - 1 :] = False
            timestamps = timestamps[~mask]
            cursor_pos = cursor_pos[~mask]
            hand_pos = hand_pos[~mask]
            eye_pos = eye_pos[~mask]

            artifact_index = np.where(timestamps[1:] < timestamps[:-1])[0]
            assert len(artifact_index) > 0

        assert len(artifact_dict["original_start_times"]) == len(artifact_index) + 1
        block_index = np.concatenate(
            [np.array([0]), artifact_index + 1, np.array([len(timestamps)])]
        )

        timestamps_blocks = [
            timestamps[block_index[i] : block_index[i + 1]]
            for i in range(len(block_index) - 1)
        ]
        for i in range(len(timestamps_blocks)):
            assert np.all(
                timestamps_blocks[i] >= artifact_dict["original_start_times"][i]
            )
            assert np.all(
                timestamps_blocks[i] <= artifact_dict["original_end_times"][i]
            )
            timestamps_blocks[i] = (
                timestamps_blocks[i]
                - artifact_dict["original_start_times"][i]
                + artifact_dict["fixed_start_times"][i]
            )

        timestamps = np.concatenate(timestamps_blocks)
        cursor_pos = np.concatenate(
            [
                cursor_pos[block_index[i] : block_index[i + 1]]
                for i in range(len(block_index) - 1)
            ]
        )
        hand_pos = np.concatenate(
            [
                hand_pos[block_index[i] : block_index[i + 1]]
                for i in range(len(block_index) - 1)
            ]
        )
        eye_pos = np.concatenate(
            [
                eye_pos[block_index[i] : block_index[i + 1]]
                for i in range(len(block_index) - 1)
            ]
        )

    # check that timestamps is sorted
    assert np.all(np.diff(timestamps) >= 0)

    # the behavior is not contiguous, so we need to determine the domain
    expected_period = 0.001  # the sampling rate of this data should be 1000 Hz
    domain_index = np.where((np.diff(timestamps) - expected_period) > 1e-4)[0]
    domain_start_index = np.insert(domain_index + 1, 0, 0)
    domain_end_index = np.append(domain_index, len(timestamps) - 1)
    domain = Interval(
        start=timestamps[domain_start_index], end=timestamps[domain_end_index] + 1e-4
    )
    assert domain.is_disjoint()
    assert domain.is_sorted()

    cursor = IrregularTimeSeries(
        timestamps=timestamps,
        pos=cursor_pos,
        domain=domain,
    )

    hand = IrregularTimeSeries(
        timestamps=timestamps,
        pos_2d=hand_pos,
        domain=domain,
    )

    eye = IrregularTimeSeries(
        timestamps=timestamps,
        pos=eye_pos,
        domain=domain,
    )

    def compute_gradient(data, timestamps, domain_start_index, domain_end_index):
        gradient = []
        # compute the velocity
        for start_index, end_index in zip(domain_start_index, domain_end_index):
            data_slice = data[start_index : end_index + 1]
            timestamps_slice = timestamps[start_index : end_index + 1]

            gradient.append(
                np.gradient(
                    data_slice,
                    timestamps_slice,
                    edge_order=1,
                    axis=0,
                )
            )
        gradient = np.concatenate(gradient)
        return gradient

    cursor.vel = compute_gradient(
        cursor.pos, timestamps, domain_start_index, domain_end_index
    )
    cursor.acc = compute_gradient(
        cursor.vel, timestamps, domain_start_index, domain_end_index
    )
    hand.vel_2d = compute_gradient(
        hand.pos_2d, timestamps, domain_start_index, domain_end_index
    )
    hand.acc_2d = compute_gradient(
        hand.vel_2d, timestamps, domain_start_index, domain_end_index
    )

    return cursor, hand, eye


def detect_outliers(cursor):
    # sometimes monkeys get angry, we want to identify the segments where the hand is
    # moving too fast, and mark them as outliers
    # we use the norm of the acceleration to identify outliers
    cursor_acc_norm = np.linalg.norm(cursor.acc, axis=1)
    mask = cursor_acc_norm > 80000.0
    # we dilate the mask to make sure we are not missing any outliers
    structure = np.ones(50, dtype=bool)
    mask = binary_dilation(mask, structure)

    # convert to interval, you need to find the start and end of the outlier segments
    start = cursor.timestamps[np.where(np.diff(mask.astype(int)) == 1)[0]]
    if mask[0]:
        start = np.insert(start, 0, cursor.timestamps[0])

    end = cursor.timestamps[np.where(np.diff(mask.astype(int)) == -1)[0]]
    if mask[-1]:
        end = np.append(end, cursor.timestamps[-1])

    cursor_outlier_segments = Interval(start=start, end=end)
    assert cursor_outlier_segments.is_disjoint()
    return cursor_outlier_segments


def extract_spikes(nwbfile, trials, artifact_dict):
    # spikes
    timestamps = []
    unit_index = []

    # units
    unit_meta = []

    electrodes = nwbfile.units.electrodes.table
    trial_table = nwbfile.trials.to_dataframe()
    unit_ctr = 0
    num_skipped_units = 0
    # all these units are obtained using threshold crossings
    for i in range(len(nwbfile.units)):
        spikes_times = nwbfile.units[i].spike_times[i]

        artifact_index = np.where(spikes_times[1:] < spikes_times[:-1])[0]
        if len(artifact_dict) > 0:
            if "skip_start" in artifact_dict:
                mask = np.logical_and(
                    spikes_times >= artifact_dict["skip_start"],
                    spikes_times <= artifact_dict["skip_end"],
                )
                if len(artifact_index) > 0:
                    mask[artifact_index[0] + 1 :] = False
                    spikes_times = spikes_times[~mask]

                artifact_index = np.where(spikes_times[1:] < spikes_times[:-1])[0]

            if len(artifact_dict["original_start_times"]) < len(artifact_index) + 1:
                logging.warning(
                    f"Unit {i} has {len(artifact_index) + 1} non-contiguous blocks, but "
                    f"found {len(artifact_dict['original_start_times'])} in the "
                    f"trials. Skipping this unit."
                )
                num_skipped_units += 1
                continue
            block_index = np.concatenate(
                [np.array([0]), artifact_index + 1, np.array([len(spikes_times)])]
            )

            spikes_times_blocks = [
                spikes_times[block_index[j] : block_index[j + 1]]
                for j in range(len(block_index) - 1)
            ]

            flag = True
            for j in range(len(spikes_times_blocks)):
                if not (
                    np.all(
                        spikes_times_blocks[j]
                        >= artifact_dict["original_start_times"][j]
                    )
                    and np.all(
                        spikes_times_blocks[j] <= artifact_dict["original_end_times"][j]
                    )
                ):
                    logging.warning(
                        f"Unit {i} has a spike time block that is not contiguous. Skipping this block."
                    )
                    num_skipped_units += 1
                    flag = False
                    break

                spikes_times_blocks[j] = (
                    spikes_times_blocks[j]
                    - artifact_dict["original_start_times"][j]
                    + artifact_dict["fixed_start_times"][j]
                )

            if not flag:
                continue

            spikes_times = np.concatenate(spikes_times_blocks)

            obs_intervals = nwbfile.units[i].obs_intervals[i]
            assert np.allclose(
                obs_intervals[:, 0], trial_table["start_time"]
            ), f"Unit {i}"
            assert np.allclose(
                obs_intervals[:, 1], trial_table["stop_time"]
            ), f"Unit {i}"

        else:
            if len(artifact_index) != 0:
                logging.warning(
                    f"Unit {i} has {len(artifact_index)} non-contiguous blocks, but no "
                    f"artifact_dict. Skipping this unit."
                )
                num_skipped_units += 1
                continue

        assert np.all(np.diff(spikes_times) >= 0)

        # label unit
        group_name = electrodes["group_name"][i]
        unit_id = f"group_{group_name}/elec{i}"

        timestamps.append(spikes_times)

        if len(spikes_times) > 0:
            unit_index.append([unit_ctr] * len(spikes_times))

        # extract unit metadata
        unit_meta.append(
            {
                "id": unit_id,
                "unit_number": unit_ctr,
                "count": len(spikes_times),
                "type": int(RecordingTech.UTAH_ARRAY_THRESHOLD_CROSSINGS),
            }
        )

        unit_ctr += 1

    if num_skipped_units > 0:
        logging.warning(
            f"Unable to resolve {num_skipped_units} out of {len(nwbfile.units)} units. "
            f"These units were excluded."
        )

    # convert unit metadata to a Data object
    unit_meta_df = pd.DataFrame(unit_meta)  # list of dicts to dataframe
    units = ArrayDict.from_dataframe(
        unit_meta_df,
        unsigned_to_long=True,
    )

    # concatenate spikes
    timestamps = np.concatenate(timestamps)
    unit_index = np.concatenate(unit_index)

    # create spikes object
    spikes = IrregularTimeSeries(
        timestamps=timestamps,
        unit_index=unit_index,
        domain=Interval(trials.start, trials.end),
    )
    spikes.sort()
    return spikes, units


def main():
    # use argparse to get arguments from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_file", type=str)
    parser.add_argument("--output_dir", type=str, default="./processed")

    args = parser.parse_args()

    brainset_description = BrainsetDescription(
        id="churchland_shenoy_neural_2012",
        origin_version="dandi/000070/draft",
        derived_version="1.0.0",
        source="https://dandiarchive.org/dandiset/000070",
        description="Monkeys recordings of Motor Cortex (M1) and dorsal Premotor Cortex"
        " (PMd) using two 96 channel high density Utah Arrays (Blackrock Microsystems) "
        "while performing reaching tasks with right hand.",
    )

    logging.info(f"Processing file: {args.input_file}")

    # open file
    io = NWBHDF5IO(args.input_file, "r")
    nwbfile = io.read()

    # extract subject metadata
    # this dataset is from dandi, which has structured subject metadata, so we
    # can use the helper function extract_subject_from_nwb
    subject = extract_subject_from_nwb(nwbfile)

    # extract experiment metadata
    recording_date = nwbfile.session_start_time.strftime("%Y%m%d")
    subject_id = subject.id
    device_id = f"{subject_id}_{recording_date}"
    session_id = f"{device_id}_center_out_reaching"

    # register session
    session_description = SessionDescription(
        id=session_id,
        recording_date=datetime.datetime.strptime(recording_date, "%Y%m%d"),
        task=Task.REACHING,
    )

    device_description = DeviceDescription(
        id=device_id,
        recording_tech=RecordingTech.UTAH_ARRAY_THRESHOLD_CROSSINGS,
    )

    # extract data about trial structure
    trials, movement_phases, artifact_dict = extract_trials(nwbfile, session_id)
    # extract spiking activity
    # this data is from dandi, we can use our helper function
    spikes, units = extract_spikes(nwbfile, trials, artifact_dict)

    # extract behavior
    cursor, hand, eye = extract_behavior(nwbfile, artifact_dict)
    assert len(cursor.domain) == len(trials)

    cursor_outlier_segments = detect_outliers(cursor)

    for key in movement_phases.keys():
        setattr(
            movement_phases,
            key,
            getattr(movement_phases, key).difference(cursor_outlier_segments),
        )

    # close file
    io.close()

    data = Data(
        brainset=brainset_description,
        subject=subject,
        session=session_description,
        device=device_description,
        # neural activity
        spikes=spikes,
        units=units,
        # stimuli and behavior
        trials=trials,
        movement_phases=movement_phases,
        cursor=cursor,
        hand=hand,
        eye=eye,
        cursor_outlier_segments=cursor_outlier_segments,
        domain=cursor.domain | spikes.domain,
    )

    # split trials into train, validation and test
    successful_trials = trials.select_by_mask(trials.is_valid)
    assert successful_trials.is_disjoint()

    train_trials, valid_trials, test_trials = successful_trials.split(
        [0.7, 0.1, 0.2], shuffle=True, random_seed=42
    )

    # we will still use the invalid trials for training
    train_trials = train_trials | trials.select_by_mask(~trials.is_valid)

    data.set_train_domain(train_trials)
    data.set_valid_domain(valid_trials)
    data.set_test_domain(test_trials)

    # save data to disk
    path = os.path.join(args.output_dir, f"{session_id}.h5")
    with h5py.File(path, "w") as file:
        data.to_hdf5(file, serialize_fn_map=serialize_fn_map)


if __name__ == "__main__":
    main()



from brainsets.core import StringIntEnum


class Cre_line(StringIntEnum):
    CUX2_CREERT2 = 0
    EMX1_IRES_CRE = 1
    FEZF2_CREER = 2
    NR5A1_CRE = 3
    NTSR1_CRE_GN220 = 4
    PVALB_IRES_CRE = 5
    RBP4_CRE_KL100 = 6
    RORB_IRES2_CRE = 7
    SCNN1A_TG3_CRE = 8
    SLC17A7_IRES2_CRE = 9
    SST_IRES_CRE = 10
    TLX3_CRE_PL56 = 11
    VIP_IRES_CRE = 12


class BrainRegion(StringIntEnum):
    VIS_RL = VISRL = 0  # Excluded
    VIS_PM = VISPM = 1
    VIS_AL = VISAL = 2
    VIS_AM = VISAM = 3
    VIS_P = VISP = 4
    VIS_L = VISL = 5



import click

from .cli_config import config
from .cli_prepare import prepare
from .cli_list import cli_list


@click.group()
def cli():
    """
    Brainsets CLI\n
    A command line interface for downloading and preparing brainsets.

    \b
    Documentation: https://brainsets.readthedocs.io/en/latest
    Project page: https://github.com/neuro-galaxy/brainsets
    """
    pass


cli.add_command(prepare)
cli.add_command(cli_list, name="list")
cli.add_command(config)


if __name__ == "__main__":
    cli()



# Taxonomy related to writing characters in cursive in the style of the Shenoy
# lab.
from enum import IntEnum, auto


class Character(IntEnum):
    a = 0
    b = auto()
    c = auto()
    d = auto()
    e = auto()
    f = auto()
    g = auto()
    h = auto()
    i = auto()
    j = auto()
    k = auto()
    l = auto()
    m = auto()
    n = auto()
    o = auto()
    p = auto()
    q = auto()
    r = auto()
    s = auto()
    t = auto()
    u = auto()
    v = auto()
    w = auto()
    x = auto()
    y = auto()
    z = auto()
    apostrophe = auto()
    comma = auto()
    doNothing = auto()
    greaterThan = auto()
    questionMark = auto()
    tilde = auto()


class Line(IntEnum):
    straightLine_short_1 = 0
    straightLine_short_2 = auto()
    straightLine_short_3 = auto()
    straightLine_short_4 = auto()
    straightLine_short_5 = auto()
    straightLine_short_6 = auto()
    straightLine_short_7 = auto()
    straightLine_short_8 = auto()
    straightLine_short_9 = auto()
    straightLine_short_10 = auto()
    straightLine_short_11 = auto()
    straightLine_short_12 = auto()
    straightLine_short_13 = auto()
    straightLine_short_14 = auto()
    straightLine_short_15 = auto()
    straightLine_short_16 = auto()
    straightLine_medium_1 = auto()
    straightLine_medium_2 = auto()
    straightLine_medium_3 = auto()
    straightLine_medium_4 = auto()
    straightLine_medium_5 = auto()
    straightLine_medium_6 = auto()
    straightLine_medium_7 = auto()
    straightLine_medium_8 = auto()
    straightLine_medium_9 = auto()
    straightLine_medium_10 = auto()
    straightLine_medium_11 = auto()
    straightLine_medium_12 = auto()
    straightLine_medium_13 = auto()
    straightLine_medium_14 = auto()
    straightLine_medium_15 = auto()
    straightLine_medium_16 = auto()
    straightLine_long_1 = auto()
    straightLine_long_2 = auto()
    straightLine_long_3 = auto()
    straightLine_long_4 = auto()
    straightLine_long_5 = auto()
    straightLine_long_6 = auto()
    straightLine_long_7 = auto()
    straightLine_long_8 = auto()
    straightLine_long_9 = auto()
    straightLine_long_10 = auto()
    straightLine_long_11 = auto()
    straightLine_long_12 = auto()
    straightLine_long_13 = auto()
    straightLine_long_14 = auto()
    straightLine_long_15 = auto()
    straightLine_long_16 = auto()



from pathlib import Path
from typing import Optional
import click
import subprocess
from prompt_toolkit import prompt

from .utils import (
    PIPELINES_PATH,
    load_config,
    get_available_brainsets,
    expand_path,
)


@click.command()
@click.argument("brainset", type=str)
@click.option("-c", "--cores", default=4, help="Number of cores to use. (default 4)")
@click.option(
    "--raw-dir",
    type=click.Path(file_okay=False),
    help="Path for storing raw data. Overrides config.",
)
@click.option(
    "--processed-dir",
    type=click.Path(file_okay=False),
    help="Path for storing processed brainset. Overrides config.",
)
@click.option(
    "--local",
    is_flag=True,
    default=False,
    help=(
        "Prepare brainset with from a local pipeline. "
        "BRAINSET must then be set to the path of the local brainset pipeline directory."
    ),
)
@click.option(
    "--use-active-env",
    is_flag=True,
    default=False,
    help=(
        "Developer flag. If set, will not create an isolated environment. "
        "Only set if you know what you're doing."
    ),
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Print debugging information.",
)
def prepare(
    brainset: str,
    cores: int,
    verbose: bool,
    use_active_env: bool,
    raw_dir: Optional[str],
    processed_dir: Optional[str],
    local: bool,
):
    """Download and process a single brainset.

    Run 'brainsets list' to get a list of available brainsets.

    \b
    Examples:
    $ brainsets prepare pei_pandarinath_nlb_2021
    $ brainsets prepare pei_pandarinath_nlb_2021 --cores 8 --raw-dir ~/data/raw --processed-dir ~/data/processed
    $ brainsets prepare ./my_local_brainsets_pipeline --local
    """

    # Get raw and processed dirs
    if raw_dir is None or processed_dir is None:
        config = load_config()
        raw_dir = expand_path(raw_dir or config["raw_dir"])
        processed_dir = expand_path(processed_dir or config["processed_dir"])
    else:
        raw_dir = expand_path(raw_dir)
        processed_dir = expand_path(processed_dir)

    if not local:
        # Preparing using an OG pipeline
        available_brainsets = get_available_brainsets()
        if brainset not in available_brainsets:
            raise click.ClickException(
                f"Brainset '{brainset}' not found. "
                f"Run 'brainsets list' to get the available list of brainsets."
            )
        # Find snakefile
        snakefile_filepath = PIPELINES_PATH / brainset / "Snakefile"
        reqs_filepath = PIPELINES_PATH / brainset / "requirements.txt"

        _validate_snakefile(snakefile_filepath)
        click.echo(f"Preparing {brainset}...")
    else:
        # Preparing using a local pipeline
        pipeline_dir = expand_path(brainset)
        snakefile_filepath = pipeline_dir / "Snakefile"
        reqs_filepath = pipeline_dir / "requirements.txt"

        _validate_snakefile(snakefile_filepath)
        click.echo(f"Preparing local pipeline: {pipeline_dir}")

    click.echo(f"Raw data directory: {raw_dir}")
    click.echo(f"Processed data directory: {processed_dir}")

    # Construct base Snakemake command with configuration
    command = [
        "snakemake",
        "-s",
        str(snakefile_filepath),
        "--config",
        f"RAW_DIR={raw_dir}",
        f"PROCESSED_DIR={processed_dir}",
        f"-c{cores}",
        "all",
        "--verbose" if verbose else "--quiet",
    ]

    if use_active_env:
        click.echo(
            "WARNING: Working in active environment due to --use-active-env.\n"
            "         This mode is only intended for brainset development purposes."
        )
        if reqs_filepath.exists():
            click.echo(
                f"WARNING: {reqs_filepath} found.\n"
                f"         These will not be installed automatically due to --use-active-env usage.\n"
                f"         Make sure to install necessary requirements manually."
            )
    elif reqs_filepath.exists():
        # If dataset has additional requirements, prefix command with uv package manager
        if not use_active_env:
            uv_prefix_command = [
                "uv",
                "run",
                "--with-requirements",
                str(reqs_filepath),
                "--isolated",
                "--no-project",
            ]
            if verbose:
                uv_prefix_command.append("--verbose")

            command = uv_prefix_command + command
            click.echo(
                "Building temporary virtual environment using"
                f" requirements from {reqs_filepath}"
            )

    # Run snakemake workflow for dataset download with live output
    try:
        process = subprocess.run(
            command,
            check=True,
            capture_output=False,
            text=True,
        )

        if process.returncode == 0:
            click.echo(f"Successfully downloaded {brainset}")
        else:
            click.echo("Error downloading dataset")
    except subprocess.CalledProcessError as e:
        click.echo(f"Error: Command failed with return code {e.returncode}")
    except Exception as e:
        click.echo(f"Error: {str(e)}")


def _validate_snakefile(filepath: Path) -> bool:

    # Check if Snakefile exists
    if not filepath.exists():
        raise click.ClickException(
            f"Missing {filepath}. A pipeline must have a Snakefile."
        )

    # Check if rule "all" exists in the Snakefile
    try:
        result = subprocess.run(
            [
                "snakemake",
                "-s",
                str(filepath),
                "--list-target-rules",
                "--config",
                "RAW_DIR=test",
                "PROCESSED_DIR=test",
            ],
            capture_output=True,
            text=True,
            check=True,
        )
        if "all" not in result.stdout.splitlines():
            raise click.ClickException(
                f"Rule 'all' not found in {filepath}. "
                " A valid Snakefile must have an 'all' rule."
            )
    except subprocess.CalledProcessError as e:
        raise click.ClickException(f"Error validating Snakefile: {e}")
